//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Detector/DetectorItem.cpp
//! @brief     Implements classes DetectorItems
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Detector/DetectorItem.h"
#include "Device/Detector/IDetector.h"
#include "Device/Mask/IShape2D.h"
#include "Device/Resolution/ResolutionFunction2DGaussian.h"
#include "GUI/Model/Detector/ResolutionFunctionItems.h"

namespace {
namespace Tag {

const QString Alignment("Alignment");
const QString DirectionVector("DirectionVector");
const QString Distance("Distance");
const QString ExpandResolutionFuncGroupbox("ExpandResolutionFuncGroupbox");
const QString Height("Height");
const QString NormalVector("NormalVector");
const QString ResolutionFunction("ResolutionFunction");
const QString Width("Width");
const QString XSize("XSize");
const QString YSize("YSize");
const QString u0("u0");
const QString v0("v0");

} // namespace Tag

const double default_detector_width = 20.0;
const double default_detector_height = 20.0;
const double default_detector_distance = 1000.0;

//! Holds the alignments as expected on the UI (map is sorted by keys)
const QMap<FlatDetector::NominalNormal, QString> alignment_names_map = {
    {FlatDetector::X, "Perpendicular to sample x-axis"},
    {FlatDetector::T, "Perpendicular to direct beam"},
    {FlatDetector::R, "Perpendicular to reflected beam"}};

} // namespace

DetectorItem::DetectorItem()
{
    m_resolutionFunction.init("Resolution function", "Detector resolution function");

    m_xSize = 100;
    m_ySize = 100;
    m_width.init("Width (mm)", "Width of the detector", default_detector_width, 3,
                 RealLimits::positive(), "width");
    m_height.init("Height (mm)", "Height of the detector", default_detector_height, 3,
                  RealLimits::positive(), "height");

    m_normalVector.init(
        "Normal vector",
        "Normal of the detector plane with length equal to the sample detector distance",
        "normalVector");
    m_normalVector.setX(default_detector_distance);
    m_directionVector.init("GisasDirection vector",
                           "Detector axis direction vector w.r.t. the sample coordinate system",
                           "directionVector");
    m_directionVector.setY(-1.0);

    m_u0.init("u0 (mm)", "", default_detector_width / 2., 3, RealLimits::limitless(), "u0");
    m_v0.init("v0 (mm)", "", 0.0, 3, RealLimits::limitless(), "v0");
    m_distance.init("Distance (mm)", "Distance from the sample origin to the detector plane",
                    default_detector_distance, "distance");

    m_detectorAlignment = ComboProperty::fromList(alignment_names_map.values(),
                                                  alignment_names_map.value(FlatDetector::T));

    updateTooltips();
}

DetectorItem::~DetectorItem() = default;

std::unique_ptr<IDetector> DetectorItem::createDetector(const Beam& beam) const
{
    auto* det = new FlatDetector(xSize(), ySize(), width(), height(), beam, detectorAlignment(),
                                 m_distance, m_u0 - width() / 2, m_v0 - height() / 2, 0, 0, 0);

    addMasksToCore(det);

    if (auto resFunc = createResolutionFunction())
        det->setResolutionFunction(*resFunc);

    return std::unique_ptr<IDetector>(det);
}

void DetectorItem::importMasks(const MaskContainerItem* maskContainer)
{
    m_maskContainerItem.copyFrom(maskContainer);
}

MaskContainerItem& DetectorItem::maskItems()
{
    return m_maskContainerItem;
}

SelectionProperty<ResolutionFunctionItemCatalog>& DetectorItem::resolutionFunctionSelection()
{
    return m_resolutionFunction;
}

std::unique_ptr<IResolutionFunction2D> DetectorItem::createResolutionFunction() const
{
    return m_resolutionFunction->createResolutionFunction(axesToCoreUnitsFactor());
}

void DetectorItem::addMasksToCore(IDetector* detector) const
{
    const double scale = axesToCoreUnitsFactor();
    ASSERT(scale == 1); // TODO remove either this or the above
    const QVector<MaskItem*> maskItems = m_maskContainerItem.maskItems();
    for (auto maskIter = maskItems.rbegin(); maskIter != maskItems.rend(); maskIter++)
        if ((*maskIter)->isVisible()) {
            if (auto* roiItem = dynamic_cast<RegionOfInterestItem*>(*maskIter)) {
                const double xlow = scale * roiItem->xLow();
                const double ylow = scale * roiItem->yLow();
                const double xup = scale * roiItem->xUp();
                const double yup = scale * roiItem->yUp();
                detector->setRegionOfInterest(xlow, ylow, xup, yup);
            } else {
                std::unique_ptr<IShape2D> shape((*maskIter)->createShape(scale));
                detector->addBinMask(*shape, (*maskIter)->maskValue());
            }
        }
}

void DetectorItem::writeTo(QXmlStreamWriter* w) const
{
    XML::writeAttribute(w, XML::Attrib::version, uint(2));

    // resolution function
    w->writeStartElement(Tag::ResolutionFunction);
    m_resolutionFunction.writeTo(w);
    w->writeEndElement();

    // resolution function groupbox: is expanded?
    w->writeStartElement(Tag::ExpandResolutionFuncGroupbox);
    XML::writeAttribute(w, XML::Attrib::value, expandResolutionFunc);
    w->writeEndElement();

    // x size
    w->writeStartElement(Tag::XSize);
    XML::writeAttribute(w, XML::Attrib::value, m_xSize);
    w->writeEndElement();

    // y size
    w->writeStartElement(Tag::YSize);
    XML::writeAttribute(w, XML::Attrib::value, m_ySize);
    w->writeEndElement();

    // width
    w->writeStartElement(Tag::Width);
    m_width.writeTo(w);
    w->writeEndElement();

    // height
    w->writeStartElement(Tag::Height);
    m_height.writeTo(w);
    w->writeEndElement();

    // alignment
    w->writeStartElement(Tag::Alignment);
    m_detectorAlignment.writeTo(w);
    w->writeEndElement();

    // normal vector
    w->writeStartElement(Tag::NormalVector);
    m_normalVector.writeTo(w);
    w->writeEndElement();

    // direction vector
    w->writeStartElement(Tag::DirectionVector);
    m_directionVector.writeTo(w);
    w->writeEndElement();

    // u0
    w->writeStartElement(Tag::u0);
    m_u0.writeTo(w);
    w->writeEndElement();

    // v0
    w->writeStartElement(Tag::v0);
    m_v0.writeTo(w);
    w->writeEndElement();

    // distance
    w->writeStartElement(Tag::Distance);
    m_distance.writeTo(w);
    w->writeEndElement();
}

void DetectorItem::readFrom(QXmlStreamReader* r)
{
    const uint version = XML::readUIntAttribute(r, XML::Attrib::version);
    Q_UNUSED(version)

    while (r->readNextStartElement()) {
        QString tag = r->name().toString();

        // resolution function
        if (tag == Tag::ResolutionFunction) {
            m_resolutionFunction.readFrom(r);
            XML::gotoEndElementOfTag(r, tag);

            // resolution function groupbox: is expanded?
        } else if (tag == Tag::ExpandResolutionFuncGroupbox) {
            XML::readAttribute(r, XML::Attrib::value, &expandResolutionFunc);
            XML::gotoEndElementOfTag(r, tag);

            // x size
        } else if (tag == Tag::XSize) {
            XML::readAttribute(r, XML::Attrib::value, &m_xSize);
            XML::gotoEndElementOfTag(r, tag);

            // y size
        } else if (tag == Tag::YSize) {
            XML::readAttribute(r, XML::Attrib::value, &m_ySize);
            XML::gotoEndElementOfTag(r, tag);

            // width
        } else if (tag == Tag::Width) {
            m_width.readFrom(r);
            XML::gotoEndElementOfTag(r, tag);

            // height
        } else if (tag == Tag::Height) {
            m_height.readFrom(r);
            XML::gotoEndElementOfTag(r, tag);

            // alignment
        } else if (tag == Tag::Alignment) {
            m_detectorAlignment.readFrom(r);
            XML::gotoEndElementOfTag(r, tag);

            // normal vector
        } else if (tag == Tag::NormalVector) {
            m_normalVector.readFrom(r);
            XML::gotoEndElementOfTag(r, tag);

            // direction vector
        } else if (tag == Tag::DirectionVector) {
            m_directionVector.readFrom(r);
            XML::gotoEndElementOfTag(r, tag);

            // u0
        } else if (tag == Tag::u0) {
            m_u0.readFrom(r);
            XML::gotoEndElementOfTag(r, tag);

            // v0
        } else if (tag == Tag::v0) {
            m_v0.readFrom(r);
            XML::gotoEndElementOfTag(r, tag);

            // distance
        } else if (tag == Tag::Distance) {
            m_distance.readFrom(r);
            XML::gotoEndElementOfTag(r, tag);

        } else
            r->skipCurrentElement();
    }
}

void DetectorItem::setDetectorAlignment(const QString& alignment)
{
    ASSERT(alignment_names_map.values().contains(alignment));
    setDetectorAlignment(alignment_names_map.key(alignment));
}

void DetectorItem::setDetectorAlignment(FlatDetector::NominalNormal alignment)
{
    m_detectorAlignment.setCurrentValue(alignment_names_map.value(alignment));
    updateTooltips();
}

FlatDetector::NominalNormal DetectorItem::detectorAlignment() const
{
    ASSERT(alignment_names_map.values().contains(m_detectorAlignment.currentValue()));
    return alignment_names_map.key(m_detectorAlignment.currentValue());
}

void DetectorItem::updateTooltips()
{
    switch (detectorAlignment()) {
    case FlatDetector::X:
        m_u0.setTooltip("u-coordinate of point where sample x-axis crosses the detector");
        m_v0.setTooltip("v-coordinate of point where sample x-axis crosses the detector");
        break;
    case FlatDetector::R:
        m_u0.setTooltip("u-coordinate of point where reflected beam hits the detector");
        m_v0.setTooltip("v-coordinate of point where reflected beam hits the detector");
        break;
    case FlatDetector::T:
        m_u0.setTooltip("u-coordinate of point where direct beam hits the detector");
        m_v0.setTooltip("v-coordinate of point where direct beam hits the detector");
        break;
    }
}
