//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/Data/DataAccessWidget.cpp
//! @brief     Defines class DataAccessWidget
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2022
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/Data/DataAccessWidget.h"
#include "Base/Util/Assert.h"
#include "GUI/Model/Data/Data1DItem.h"
#include "GUI/Model/Data/Data2DItem.h"
#include "GUI/Model/Device/RealItem.h"
#include "GUI/Model/Job/JobItem.h"

DataAccessWidget::DataAccessWidget(QWidget* parent)
    : QWidget(parent)
    , m_item(nullptr)
    , m_fftItem(nullptr)
{
}
DataAccessWidget::~DataAccessWidget() = default;

void DataAccessWidget::setBaseItem(QObject* item)
{
    m_item = item;
    ASSERT(jobItem() || realItem());

    disconnect(m_item, &QObject::destroyed, this, nullptr);
    connect(m_item, &QObject::destroyed, this, [this]() { m_item = nullptr; });
}

void DataAccessWidget::setFourierItem(Data2DItem* fftItem)
{
    m_fftItem = fftItem;
}

QList<QAction*> DataAccessWidget::actionList()
{
    return QList<QAction*>();
}

JobItem* DataAccessWidget::jobItem() const
{
    return dynamic_cast<JobItem*>(m_item);
}

RealItem* DataAccessWidget::realItem() const
{
    return dynamic_cast<RealItem*>(m_item);
}

QObject* DataAccessWidget::jobRealBase() const
{
    return m_item;
}

Data1DItem* DataAccessWidget::realData1DItem() const
{
    if (jobItem()) {
        if (const RealItem* ii = jobItem()->realItem())
            return dynamic_cast<Data1DItem*>(ii->dataItem());
        return nullptr;
    }
    if (realItem())
        return dynamic_cast<Data1DItem*>(realItem()->dataItem());
    return nullptr;
}

Data2DItem* DataAccessWidget::realData2DItem() const
{
    if (jobItem()) {
        if (const RealItem* ii = jobItem()->realItem())
            return dynamic_cast<Data2DItem*>(ii->dataItem());
        return nullptr;
    }
    if (realItem())
        return dynamic_cast<Data2DItem*>(realItem()->dataItem());
    return nullptr;
}

Data1DItem* DataAccessWidget::simuData1DItem() const
{
    return jobItem() ? dynamic_cast<Data1DItem*>(jobItem()->simulatedDataItem()) : nullptr;
}

Data2DItem* DataAccessWidget::simuData2DItem() const
{
    return jobItem() ? dynamic_cast<Data2DItem*>(jobItem()->simulatedDataItem()) : nullptr;
}

Data1DItem* DataAccessWidget::diffData1DItem() const
{
    return jobItem() ? dynamic_cast<Data1DItem*>(jobItem()->diffDataItem()) : nullptr;
}

Data2DItem* DataAccessWidget::diffData2DItem() const
{
    return jobItem() ? dynamic_cast<Data2DItem*>(jobItem()->diffDataItem()) : nullptr;
}

QList<Data1DItem*> DataAccessWidget::mainData1DItems() const
{
    QList<Data1DItem*> result;
    if (auto* ii = simuData1DItem())
        result.append(ii);
    if (auto* ii = realData1DItem())
        result.append(ii);
    return result;
}

QList<Data2DItem*> DataAccessWidget::mainData2DItems() const
{
    if (m_fftItem)
        return {m_fftItem};
    QList<Data2DItem*> result;
    if (auto* ii = simuData2DItem())
        result.append(ii);
    if (auto* ii = realData2DItem())
        result.append(ii);
    return result;
}

QList<Data1DItem*> DataAccessWidget::allData1DItems() const
{
    QList<Data1DItem*> result = mainData1DItems();
    if (auto* ii = diffData1DItem())
        result.append(ii);
    return result;
}

QList<Data2DItem*> DataAccessWidget::allData2DItems() const
{
    if (m_fftItem)
        return {m_fftItem};
    QList<Data2DItem*> result = mainData2DItems();
    if (auto* ii = diffData2DItem())
        result.append(ii);
    return result;
}

Data1DItem* DataAccessWidget::currentData1DItem() const
{
    if (allData1DItems().size() == 0)
        return nullptr;
    return allData1DItems().first();
}

Data2DItem* DataAccessWidget::currentData2DItem() const
{
    if (allData2DItems().size() == 0)
        return nullptr;
    return allData2DItems().first();
}
