//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/Mask/MaskEditorCanvas.cpp
//! @brief     Implements class MaskEditorCanvas
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/Mask/MaskEditorCanvas.h"
#include "GUI/Model/Data/Data2DItem.h"
#include "GUI/Model/Data/MaskResultsPresenter.h"
#include "GUI/Model/Mask/MaskItems.h"
#include "GUI/Model/Project/ProjectDocument.h"
#include "GUI/View/Mask/MaskGraphicsScene.h"
#include "GUI/View/Mask/MaskGraphicsView.h"
#include "GUI/View/PlotScale/SavePlotAssistant.h"
#include "GUI/View/PlotUtil/ColorMap.h"
#include "GUI/View/PlotUtil/PlotStatusLabel.h"
#include <QVBoxLayout>

MaskEditorCanvas::MaskEditorCanvas(QWidget* parent)
    : QWidget(parent)
    , m_scene(new MaskGraphicsScene(this))
    , m_view(new MaskGraphicsView(m_scene))
    , m_data2DItem(nullptr)
    , m_statusLabel(new PlotStatusLabel(nullptr, this))
    , m_resultsPresenter(new MaskResultsPresenter(this))
{
    setObjectName("MaskEditorCanvas");
    setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Expanding);

    auto* mainLayout = new QVBoxLayout;
    mainLayout->addWidget(m_view);
    mainLayout->addWidget(m_statusLabel);
    mainLayout->setContentsMargins(0, 0, 0, 0);
    mainLayout->setSpacing(0);
    setLayout(mainLayout);

    connect(m_view, &MaskGraphicsView::changeActivityRequest, this,
            &MaskEditorCanvas::changeActivityRequest);
    connect(m_view, &MaskGraphicsView::deleteSelectedRequest, this,
            &MaskEditorCanvas::deleteSelectedRequest);
}

void MaskEditorCanvas::setContextMEC(Data2DItem* data2DItem)
{
    m_data2DItem = data2DItem;

    m_scene->setAssociatedItems(data2DItem, data2DItem->maskContainerItem());
    m_resultsPresenter->setContextMRP(data2DItem);
    m_view->updateSize(m_view->size());

    m_statusLabel->addPlot(m_scene->colorMap());
}

void MaskEditorCanvas::resetContextMEC()
{
    m_data2DItem = nullptr;
    m_scene->resetContextMGS();
    m_statusLabel->reset();
}

void MaskEditorCanvas::setSelectionModel(QItemSelectionModel* model)
{
    m_scene->setSelectionModel(model);
}

MaskGraphicsScene* MaskEditorCanvas::getScene()
{
    return m_scene;
}

void MaskEditorCanvas::onPresentationTypeRequest(MaskEditorFlags::PresentationType presentationType)
{
    m_scene->clearSelection(); // important to avoid crash (unsubscribe while calling subscribers)
    m_resultsPresenter->updatePresenter(presentationType);

    if (auto* container = m_data2DItem->maskContainerItem()) {
        bool tempVisible = presentationType.testFlag(MaskEditorFlags::MASK_EDITOR);
        for (MaskItem* mask : container->maskItems())
            mask->setIsVisible(tempVisible && mask->wasVisible());
    }
}

//! Saves plot into project directory.

void MaskEditorCanvas::onSavePlotRequest()
{
    ASSERT(gProjectDocument.has_value());
    QString dirname = gProjectDocument.value()->userExportDir();
    GUI::Plot::savePlot(dirname, m_scene->colorMap()->customPlot(), m_data2DItem->c_field());
}

void MaskEditorCanvas::onResetViewRequest()
{
    m_view->onResetViewRequest();

    if (m_data2DItem->isZoomed())
        m_data2DItem->resetView();
    else
        setZoomToROI();
    gProjectDocument.value()->setModified();
}

void MaskEditorCanvas::setZoomToROI()
{
    if (MaskContainerItem* maskContainer = m_data2DItem->maskContainerItem()) {
        if (auto* roiItem = maskContainer->regionOfInterestItem()) {
            m_data2DItem->setXrange(roiItem->xLow(), roiItem->xUp());
            m_data2DItem->setYrange(roiItem->yLow(), roiItem->yUp());
        }
    }
}
